import copy

import dask
import pandas as pd
from intake.source.base import DataSource, Schema

from .merge_util import _aggregate, _open_asset, _path_to_mapper, _to_nested_dict

_DATA_FORMAT_KEY = '_data_format_'


class ESMDataSource(DataSource):
    version = '1.0'
    container = 'xarray'
    name = 'esm_single_source'
    partition_access = True

    def __init__(
        self,
        key,
        row,
        path_column,
        data_format=None,
        format_column=None,
        cdf_kwargs=None,
        zarr_kwargs=None,
        storage_options=None,
        preprocess=None,
        requested_variables=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.key = key
        self.cdf_kwargs = cdf_kwargs or {'chunks': {}}
        self.zarr_kwargs = zarr_kwargs or {}
        self.storage_options = storage_options or {}
        self.preprocess = preprocess
        self.requested_variables = requested_variables
        if not isinstance(row, pd.Series) or row.empty:
            raise ValueError('`row` must be a non-empty pandas.Series')
        self.row = row.copy()
        self.path_column = path_column
        self._ds = None
        if format_column is not None:
            self.data_format = self.row[format_column]
        elif data_format:
            self.data_format = data_format
        else:
            raise ValueError('Please specify either `data_format` or `format_column`')

    def __repr__(self):
        return f'<name: {self.key}, asset: 1'

    def _get_schema(self):

        if self._ds is None:
            self._open_dataset()

            metadata = {
                'dims': dict(self._ds.dims),
                'data_vars': {k: list(self._ds[k].coords) for k in self._ds.data_vars.keys()},
                'coords': tuple(self._ds.coords.keys()),
            }
            self._schema = Schema(
                datashape=None,
                dtype=None,
                shape=None,
                npartitions=None,
                extra_metadata=metadata,
            )
        return self._schema

    def _open_dataset(self):
        mapper = _path_to_mapper(self.row[self.path_column], self.storage_options, self.data_format)
        ds = _open_asset(
            mapper,
            data_format=self.data_format,
            zarr_kwargs=self.zarr_kwargs,
            cdf_kwargs=self.cdf_kwargs,
            preprocess=self.preprocess,
            requested_variables=self.requested_variables,
        )
        ds.attrs['intake_esm_dataset_key'] = self.key
        self._ds = ds
        return ds

    def to_dask(self):
        """Return xarray object (which will have chunks)"""
        self._load_metadata()
        return self._ds

    def close(self):
        """Delete open files from memory"""
        self._ds = None
        self._schema = None


class ESMGroupDataSource(DataSource):
    version = '1.0'
    container = 'xarray'
    name = 'esm_group'
    partition_access = True

    def __init__(
        self,
        key,
        df,
        aggregation_dict,
        path_column,
        variable_column,
        data_format=None,
        format_column=None,
        cdf_kwargs=None,
        zarr_kwargs=None,
        storage_options=None,
        preprocess=None,
        requested_variables=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.key = key
        self.cdf_kwargs = cdf_kwargs or {'chunks': {}}
        self.zarr_kwargs = zarr_kwargs or {}
        self.storage_options = storage_options or {}
        self.preprocess = preprocess
        self.requested_variables = requested_variables
        self._ds = None
        if not isinstance(df, pd.DataFrame) or df.empty:
            raise ValueError('`df` must be a non-empty pandas.DataFrame')
        self.df = df.copy()
        self.aggregation_columns, self.aggregation_dict = _sanitize_aggregations(
            df, aggregation_dict
        )
        self.path_column = path_column
        self.variable_column = variable_column
        if format_column is not None:
            self.df[_DATA_FORMAT_KEY] = df[format_column]
        else:
            if data_format is None:
                raise ValueError('Please specify either `data_format` or `format_column`')
            self.df[_DATA_FORMAT_KEY] = [data_format] * len(df)

        self.data_format = data_format
        self.format_column = format_column

    def __repr__(self):
        """Make string representation of object."""
        contents = f'<name: {self.name}, assets: {len(self.df)}'
        return contents

    def _ipython_display_(self):
        """
        Display the entry as a rich object in an IPython session
        """
        from IPython.display import HTML, display

        columns = list(set([self.path_column, self.variable_column] + self.aggregation_columns))
        text = self.df[columns].to_html()
        contents = f"""
        <p>
            <ul>
                <li><strong>Name</strong>                 : {self.name}</li>
                <li><strong>Num of xarray.Dataset</strong>: 1</li>
                <li><strong>Num of assets</strong>        : {len(self.df)}</li>
                <li><strong>Aggregation columns</strong>  : {str(self.aggregation_columns)}</li>
            </ul>
        </p>
        {text}
        """
        display(HTML(contents))

    def _get_schema(self):

        if self._ds is None:
            self._open_dataset()

            metadata = {
                'dims': dict(self._ds.dims),
                'data_vars': {k: list(self._ds[k].coords) for k in self._ds.data_vars.keys()},
                'coords': tuple(self._ds.coords.keys()),
            }
            self._schema = Schema(
                datashape=None,
                dtype=None,
                shape=None,
                npartitions=None,
                extra_metadata=metadata,
            )
        return self._schema

    def _open_dataset(self):
        @dask.delayed
        def read_dataset(
            path,
            data_format,
            storage_options,
            cdf_kwargs,
            zarr_kwargs,
            preprocess=None,
            varname=None,
        ):
            # replace path column with mapper (dependent on filesystem type)
            mapper = _path_to_mapper(path, storage_options, data_format)
            ds = _open_asset(
                mapper,
                data_format=data_format,
                zarr_kwargs=zarr_kwargs,
                cdf_kwargs=cdf_kwargs,
                preprocess=preprocess,
                varname=varname,
                requested_variables=self.requested_variables,
            )
            return (path, ds)

        datasets = [
            read_dataset(
                row[self.path_column],
                row[_DATA_FORMAT_KEY],
                self.storage_options,
                self.cdf_kwargs,
                self.zarr_kwargs,
                self.preprocess,
                row[self.variable_column],
            )
            for _, row in self.df.iterrows()
        ]
        datasets = dask.compute(*datasets)
        mapper_dict = dict(datasets)
        nd = create_nested_dict(self.df, self.path_column, self.aggregation_columns)
        n_agg = len(self.aggregation_columns)

        ds = _aggregate(
            self.aggregation_dict,
            self.aggregation_columns,
            n_agg,
            nd,
            mapper_dict,
            self.key,
        )
        ds.attrs['intake_esm_dataset_key'] = self.key
        self._ds = ds
        return ds

    def to_dask(self):
        """Return xarray object (which will have chunks)"""
        self._load_metadata()
        return self._ds

    def close(self):
        """Delete open files from memory"""
        self._ds = None
        self._schema = None


def create_nested_dict(df, path_column, aggregation_columns):
    mi = df.set_index(aggregation_columns)
    nd = _to_nested_dict(mi[path_column])
    return nd


def _sanitize_aggregations(df, aggregation_dict):
    _aggregation_dict = copy.deepcopy(aggregation_dict)
    agg_columns = list(_aggregation_dict.keys())
    drop_cols = []
    for col in agg_columns:
        if df[col].isnull().all():
            drop_cols.append(col)
            del _aggregation_dict[col]
        elif df[col].isnull().any():
            raise ValueError(
                f'The data in the {col} column should either be all NaN or there should be no NaNs'
            )

    aggregation_columns = list(filter(lambda x: x not in drop_cols, agg_columns))

    return aggregation_columns, _aggregation_dict



import concurrent.futures
import json
import pathlib
from collections import OrderedDict, namedtuple
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Union
from warnings import warn

import dask
import intake
import pandas as pd
import xarray as xr
from fastprogress.fastprogress import progress_bar
from intake.catalog import Catalog

from .search import _get_columns_with_iterables, _unique, search
from .utils import _fetch_and_parse_json, _fetch_catalog

_AGGREGATIONS_TYPES = {'join_existing', 'join_new', 'union'}


class esm_datastore(Catalog):
    """
    An intake plugin for parsing an ESM (Earth System Model) Collection/catalog
    and loading assets (netCDF files and/or Zarr stores) into xarray datasets.
    The in-memory representation for the catalog is a Pandas DataFrame.

    Parameters
    ----------
    esmcol_obj : str, pandas.DataFrame
        If string, this must be a path or URL to an ESM collection JSON file.
        If pandas.DataFrame, this must be the catalog content that would otherwise
        be in a CSV file.
    esmcol_data : dict, optional
            ESM collection spec information, by default None
    progressbar : bool, optional
        Will print a progress bar to standard error (stderr)
        when loading assets into :py:class:`~xarray.Dataset`,
        by default True
    sep : str, optional
        Delimiter to use when constructing a key for a query, by default '.'
    csv_kwargs : dict, optional
        Additional keyword arguments passed through to the
        :py:func:`~pandas.read_csv` function.
    **kwargs :
        Additional keyword arguments are passed through to the
        :py:class:`~intake.catalog.Catalog` base class.

    Examples
    --------

    At import time, this plugin is available in intake's registry as `esm_datastore` and
    can be accessed with `intake.open_esm_datastore()`:

    >>> import intake
    >>> url = "https://raw.githubusercontent.com/NCAR/intake-esm-datastore/master/catalogs/pangeo-cmip6.json"
    >>> col = intake.open_esm_datastore(url)
    >>> col.df.head()
    activity_id institution_id source_id experiment_id  ... variable_id grid_label                                             zstore dcpp_init_year
    0  AerChemMIP            BCC  BCC-ESM1        ssp370  ...          pr         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    1  AerChemMIP            BCC  BCC-ESM1        ssp370  ...        prsn         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    2  AerChemMIP            BCC  BCC-ESM1        ssp370  ...         tas         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    3  AerChemMIP            BCC  BCC-ESM1        ssp370  ...      tasmax         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    4  AerChemMIP            BCC  BCC-ESM1        ssp370  ...      tasmin         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    """

    name = 'esm_datastore'
    container = 'xarray'

    def __init__(
        self,
        esmcol_obj: Union[str, pd.DataFrame],
        esmcol_data: Dict[str, Any] = None,
        progressbar: bool = True,
        sep: str = '.',
        csv_kwargs: Dict[str, Any] = None,
        **kwargs,
    ):

        """Intake Catalog representing an ESM Collection."""
        super(esm_datastore, self).__init__(**kwargs)
        if isinstance(esmcol_obj, (str, pathlib.PurePath)):
            self.esmcol_data, self.esmcol_path = _fetch_and_parse_json(esmcol_obj)
            self._df, self.catalog_file = _fetch_catalog(self.esmcol_data, esmcol_obj, csv_kwargs)

        elif isinstance(esmcol_obj, pd.DataFrame):
            if esmcol_data is None:
                raise ValueError("Missing required argument: 'esmcol_data'")
            self._df = esmcol_obj
            self.esmcol_data = esmcol_data
            self.esmcol_path = None
            self.catalog_file = None
        else:
            raise ValueError(
                f'{self.name} constructor not properly called! `esmcol_obj` is of type: {type(esmcol_obj)}, however valid types of `esmcol_obj` are either `str` or `pathlib.PurePath` or `pandas.DataFrame`. '
            )

        self.progressbar = progressbar
        self._kwargs = kwargs
        self._to_dataset_args_token = None
        self._datasets = None
        self.sep = sep
        self._data_format, self._format_column_name = None, None
        self._path_column_name = self.esmcol_data['assets']['column_name']
        if 'format' in self.esmcol_data['assets']:
            self._data_format = self.esmcol_data['assets']['format']
        else:
            self._format_column_name = self.esmcol_data['assets']['format_column_name']
        self._columns_with_iterables = _get_columns_with_iterables(self.df)
        self.aggregation_info = self._get_aggregation_info()
        self._entries = {}
        self._set_groups_and_keys()
        self._requested_variables = []

        if self.variable_column_name:
            self._multiple_variable_assets = (
                self.variable_column_name in self._columns_with_iterables
            )
        else:
            self._multiple_variable_assets = False

    def _set_groups_and_keys(self):
        if self.aggregation_info.groupby_attrs and set(self.df.columns) != set(
            self.aggregation_info.groupby_attrs
        ):
            self._grouped = self.df.groupby(self.aggregation_info.groupby_attrs)
            internal_keys = self._grouped.groups.keys()
            public_keys = []
            for key in internal_keys:
                if isinstance(key, str):
                    p_key = key
                else:
                    p_key = self.sep.join(str(v) for v in key)
                public_keys.append(p_key)

        else:
            self._grouped = self.df
            internal_keys = list(self._grouped.index)
            public_keys = [
                self.sep.join(str(v) for v in row.values) for _, row in self._grouped.iterrows()
            ]

        self._keys = dict(zip(public_keys, internal_keys))

    def _allnan_or_nonan(self, column: str) -> bool:
        """
        Helper function used to filter groupby_attrs to ensure no columns with all nans

        Parameters
        ----------
        column : str
            Column name

        Returns
        -------
        bool
            Whether the dataframe column has all NaNs or no NaN valles

        Raises
        ------
        ValueError
            When the column has a mix of NaNs non NaN values
        """
        if self.df[column].isnull().all():
            return False
        if self.df[column].isnull().any():
            raise ValueError(
                f'The data in the {column} column should either be all NaN or there should be no NaNs'
            )
        return True

    def _get_aggregation_info(self):

        AggregationInfo = namedtuple(
            'AggregationInfo',
            [
                'groupby_attrs',
                'variable_column_name',
                'aggregations',
                'agg_columns',
                'aggregation_dict',
            ],
        )

        groupby_attrs = []
        variable_column_name = None
        aggregations = []
        aggregation_dict = {}
        agg_columns = []

        if 'aggregation_control' in self.esmcol_data:
            variable_column_name = self.esmcol_data['aggregation_control']['variable_column_name']
            groupby_attrs = self.esmcol_data['aggregation_control'].get('groupby_attrs', [])
            aggregations = self.esmcol_data['aggregation_control'].get('aggregations', [])
            aggregations, aggregation_dict, agg_columns = _construct_agg_info(aggregations)
            groupby_attrs = list(filter(self._allnan_or_nonan, groupby_attrs))

        if not aggregations:
            groupby_attrs = []

        # Cast all agg_columns with iterables to tuple values so as
        # to avoid hashing issues (e.g. TypeError: unhashable type: 'list')

        columns = set(self._columns_with_iterables).intersection(set(agg_columns))
        if columns:
            for column in columns:
                self.df[column] = self.df[column].map(tuple)

        aggregation_info = AggregationInfo(
            groupby_attrs,
            variable_column_name,
            aggregations,
            agg_columns,
            aggregation_dict,
        )
        return aggregation_info

    def keys(self) -> List:
        """
        Get keys for the catalog entries

        Returns
        -------
        list
            keys for the catalog entries
        """
        return self._keys.keys()

    @property
    def key_template(self) -> str:
        """
        Return string template used to create catalog entry keys

        Returns
        -------
        str
          string template used to create catalog entry keys
        """
        if self.aggregation_info.groupby_attrs:
            template = self.sep.join(self.aggregation_info.groupby_attrs)
        else:
            template = self.sep.join(self.df.columns)
        return template

    @property
    def df(self) -> pd.DataFrame:
        """
        Return pandas :py:class:`~pandas.DataFrame`.
        """
        return self._df

    @df.setter
    def df(self, value: pd.DataFrame):
        self._df = value
        self._set_groups_and_keys()

    @property
    def groupby_attrs(self) -> list:
        """
        Dataframe columns used to determine groups of compatible datasets.

        Returns
        -------
        list
            Columns used to determine groups of compatible datasets.
        """
        return self.aggregation_info.groupby_attrs

    @groupby_attrs.setter
    def groupby_attrs(self, value: list) -> None:
        groupby_attrs = list(filter(self._allnan_or_nonan, value))
        self.aggregation_info = self.aggregation_info._replace(groupby_attrs=groupby_attrs)
        self._set_groups_and_keys()
        self._entries = {}

    @property
    def variable_column_name(self) -> str:
        """
        Name of the column that contains the variable name.
        """
        return self.aggregation_info.variable_column_name

    @variable_column_name.setter
    def variable_column_name(self, value: str) -> None:
        self.aggregation_info = self.aggregation_info._replace(variable_column_name=value)

    @property
    def aggregations(self):
        return self.aggregation_info.aggregations

    @property
    def agg_columns(self) -> list:
        """
        List of columns used to merge/concatenate compatible
        multiple :py:class:`~xarray.Dataset` into a single :py:class:`~xarray.Dataset`.
        """
        return self.aggregation_info.agg_columns

    @property
    def aggregation_dict(self) -> dict:
        return self.aggregation_info.aggregation_dict

    def update_aggregation(
        self, attribute_name: str, agg_type: str = None, options: dict = None, delete=False
    ):
        """
        Updates aggregation operations info.

        Parameters
        ----------
        attribute_name : str
            Name of attribute (column) across which to aggregate.

        agg_type : str, optional
            Type of aggregation operation to apply. Valid values include:
            `join_new`, `join_existing`, `union`, by default None

        options : dict, optional
            Aggregration settings that are passed as keywords arguments to
            :py:func:`~xarray.concat` or :py:func:`~xarray.merge`. For `join_existing`, it must contain
            the name of the existing dimension to use (for e.g.: something like {'dim': 'time'}).,
            by default None

        delete : bool, optional
             Whether to delete/remove/disable aggregation operations for a particular attribute,
             by default False
        """

        def validate_type(t):
            assert (
                t in _AGGREGATIONS_TYPES
            ), f'Invalid aggregation agg_type={t}. Valid values are: {list(_AGGREGATIONS_TYPES)}.'

        def validate_attribute_name(name):
            assert (
                name in self.df.columns
            ), f'Attribute_name={attribute_name} is invalid. Attribute name must exist as a column in the dataframe. Valid values: {self.df.columns.tolist()}.'

        def validate_options(options):
            assert isinstance(
                options, dict
            ), f'Options must be a dictionary. Found the type of options={options} to be {type(options)}.'

        aggregations = self.aggregations.copy()
        validate_attribute_name(attribute_name)
        found = False
        match = None
        idx = None
        for index, agg in enumerate(aggregations):
            if agg['attribute_name'] == attribute_name:
                found = True
                match = agg
                idx = index
                break

        if found:
            if delete:
                del aggregations[idx]
            else:
                if agg_type is not None:
                    validate_type(agg_type)
                    match['type'] = agg_type
                if options is not None:
                    validate_options(options)
                    match['options'] = options
                aggregations[idx] = match

        else:
            if delete:
                message = f'No change. Tried removing/deleting/disabling non-existing aggregation operations for attribute={attribute_name}'
                warn(message)
            else:
                match = {}
                validate_type(agg_type)
                match['type'] = agg_type
                match['attribute_name'] = attribute_name
                if options is not None:
                    validate_options(options)
                    match['options'] = options
                elif options is None:
                    match['options'] = {}
                aggregations.append(match)

        aggregations, aggregation_dict, agg_columns = _construct_agg_info(aggregations)
        kwargs = {
            'aggregations': aggregations,
            'aggregation_dict': aggregation_dict,
            'agg_columns': agg_columns,
        }
        if len(aggregations) == 0:
            warn(
                'Setting `groupby_attrs` to []. Aggregations will be disabled because `groupby_attrs` is empty.'
            )
            kwargs['groupby_attrs'] = []
        self.aggregation_info = self.aggregation_info._replace(**kwargs)
        self._entries = {}
        if len(self.groupby_attrs) == 0:
            self._set_groups_and_keys()

    @property
    def path_column_name(self) -> str:
        """
        The name of the column containing the path to the asset.
        """
        return self._path_column_name

    @path_column_name.setter
    def path_column_name(self, value: str) -> None:
        self._path_column_name = value

    @property
    def data_format(self) -> str:
        """
        The data format. Valid values are netcdf and zarr.
        If specified, it means that all data assets in the catalog use the same data format.
        """
        return self._data_format

    @data_format.setter
    def data_format(self, value: str) -> None:
        self._data_format = value

    @property
    def format_column_name(self) -> str:
        """
        Name of the column which contains the data format.
        """
        return self._format_column_name

    @format_column_name.setter
    def format_column_name(self, value: str) -> None:
        self._format_column_name = value

    def __len__(self):
        return len(self.keys())

    def _get_entries(self):
        # Due to just-in-time entry creation, we may not have all entries loaded
        # We need to make sure to create entries missing from self._entries
        missing = set(self.keys()) - set(self._entries.keys())
        for key in missing:
            _ = self[key]
        return self._entries

    def __getitem__(self, key: str):
        """
        This method takes a key argument and return a data source
        corresponding to assets (files) that will be aggregated into a
        single xarray dataset.

        Parameters
        ----------
        key : str
          key to use for catalog entry lookup

        Returns
        -------
        intake_esm.source.ESMGroupDataSource
             A data source by name (key)

        Raises
        ------
        KeyError
            if key is not found.

        Examples
        --------
        >>> col = intake.open_esm_datastore("mycatalog.json")
        >>> data_source = col["AerChemMIP.BCC.BCC-ESM1.piClim-control.AERmon.gn"]
        """
        # The canonical unique key is the key of a compatible group of assets
        try:
            return self._entries[key]
        except KeyError:
            if key in self.keys():
                internal_key = self._keys[key]
                if isinstance(self._grouped, pd.DataFrame):
                    df = self._grouped.loc[internal_key]
                    args = dict(
                        key=key,
                        row=df,
                        path_column=self.path_column_name,
                        data_format=self.data_format,
                        format_column=self.format_column_name,
                        requested_variables=self._requested_variables,
                    )
                    entry = _make_entry(key, 'esm_single_source', args)
                else:
                    df = self._grouped.get_group(internal_key)
                    args = dict(
                        df=df,
                        aggregation_dict=self.aggregation_info.aggregation_dict,
                        path_column=self.path_column_name,
                        variable_column=self.aggregation_info.variable_column_name,
                        data_format=self.data_format,
                        format_column=self.format_column_name,
                        key=key,
                        requested_variables=self._requested_variables,
                    )
                    entry = _make_entry(key, 'esm_group', args)

                self._entries[key] = entry
                return self._entries[key]
            raise KeyError(key)

    def __contains__(self, key):
        # Python falls back to iterating over the entire catalog
        # if this method is not defined. To avoid this, we implement it differently

        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def __repr__(self):
        """Make string representation of object."""
        return f'<{self.esmcol_data["id"]} catalog with {len(self)} dataset(s) from {len(self.df)} asset(s)>'

    def _repr_html_(self):
        """
        Return an html representation for the catalog object.
        Mainly for IPython notebook
        """
        uniques = pd.DataFrame(self.nunique(), columns=['unique'])
        text = uniques._repr_html_()
        output = f'<p><strong>{self.esmcol_data["id"]} catalog with {len(self)} dataset(s) from {len(self.df)} asset(s)</strong>:</p> {text}'
        return output

    def _ipython_display_(self):
        """
        Display the entry as a rich object in an IPython session
        """
        from IPython.display import HTML, display

        contents = self._repr_html_()
        display(HTML(contents))

    def __dir__(self):
        rv = [
            'df',
            'to_dataset_dict',
            'from_df',
            'keys',
            'serialize',
            'search',
            'unique',
            'nunique',
            'update_aggregation',
            'key_template',
            'groupby_attrs',
            'variable_column_name',
            'aggregations',
            'agg_columns',
            'aggregation_dict',
            'path_column_name',
            'data_format',
            'format_column_name',
        ]
        return sorted(list(self.__dict__.keys()) + rv)

    def _ipython_key_completions_(self):
        return self.__dir__()

    @classmethod
    def from_df(
        cls,
        df: pd.DataFrame,
        esmcol_data: Dict[str, Any] = None,
        progressbar: bool = True,
        sep: str = '.',
        **kwargs,
    ) -> 'esm_datastore':
        """
        Create catalog from the given dataframe

        Parameters
        ----------
        df : pandas.DataFrame
            catalog content that would otherwise be in a CSV file.
        esmcol_data : dict, optional
            ESM collection spec information, by default None
        progressbar : bool, optional
            Will print a progress bar to standard error (stderr)
            when loading assets into :py:class:`~xarray.Dataset`,
            by default True
        sep : str, optional
            Delimiter to use when constructing a key for a query, by default '.'

        Returns
        -------
        :py:class:`~intake_esm.core.esm_datastore`
            Catalog object
        """
        return cls(
            df,
            esmcol_data=esmcol_data,
            progressbar=progressbar,
            sep=sep,
            **kwargs,
        )

    def search(self, require_all_on: Union[str, List] = None, **query):
        """Search for entries in the catalog.

        Parameters
        ----------
        require_all_on : list, str, optional
            A dataframe column or a list of dataframe columns across
            which all entries must satisfy the query criteria.
            If None, return entries that fulfill any of the criteria specified
            in the query, by default None.
        **query:
            keyword arguments corresponding to user's query to execute against the dataframe.

        Returns
        -------
        cat : :py:class:`~intake_esm.core.esm_datastore`
          A new Catalog with a subset of the entries in this Catalog.

        Examples
        --------
        >>> import intake
        >>> col = intake.open_esm_datastore("pangeo-cmip6.json")
        >>> col.df.head(3)
        activity_id institution_id source_id  ... grid_label                                             zstore dcpp_init_year
        0  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
        1  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
        2  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN

        >>> cat = col.search(
        ...     source_id=["BCC-CSM2-MR", "CNRM-CM6-1", "CNRM-ESM2-1"],
        ...     experiment_id=["historical", "ssp585"],
        ...     variable_id="pr",
        ...     table_id="Amon",
        ...     grid_label="gn",
        ... )
        >>> cat.df.head(3)
            activity_id institution_id    source_id  ... grid_label                                             zstore dcpp_init_year
        260        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r1i...            NaN
        346        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r2i...            NaN
        401        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r3i...            NaN

        The search method also accepts compiled regular expression objects
        from :py:func:`~re.compile` as patterns.

        >>> import re
        >>> # Let's search for variables containing "Frac" in their name
        >>> pat = re.compile(r"Frac")  # Define a regular expression
        >>> cat.search(variable_id=pat)
        >>> cat.df.head().variable_id
        0     residualFrac
        1    landCoverFrac
        2    landCoverFrac
        3     residualFrac
        4    landCoverFrac
        """

        results = search(self.df, require_all_on=require_all_on, **query)
        if self._multiple_variable_assets:
            requested_variables = query.get(self.variable_column_name, [])
        else:
            requested_variables = []
        ret = esm_datastore.from_df(
            results,
            esmcol_data=self.esmcol_data,
            progressbar=self.progressbar,
            sep=self.sep,
            **self._kwargs,
        )
        ret._requested_variables = requested_variables
        return ret

    def serialize(self, name: str, directory: str = None, catalog_type: str = 'dict') -> None:
        """Serialize collection/catalog to corresponding json and csv files.

        Parameters
        ----------
        name : str
            name to use when creating ESM collection json file and csv catalog.
        directory : str, PathLike, default None
            The path to the local directory. If None, use the current directory
        catalog_type: str, default 'dict'
            Whether to save the catalog table as a dictionary in the JSON file or as a separate CSV file.

        Notes
        -----
        Large catalogs can result in large JSON files. To keep the JSON file size manageable, call with
        `catalog_type='file'` to save catalog as a separate CSV file.

        Examples
        --------
        >>> import intake
        >>> col = intake.open_esm_datastore("pangeo-cmip6.json")
        >>> col_subset = col.search(
        ...     source_id="BCC-ESM1",
        ...     grid_label="gn",
        ...     table_id="Amon",
        ...     experiment_id="historical",
        ... )
        >>> col_subset.serialize(name="cmip6_bcc_esm1", catalog_type="file")
        Writing csv catalog to: cmip6_bcc_esm1.csv.gz
        Writing ESM collection json file to: cmip6_bcc_esm1.json
        """

        def _clear_old_catalog(catalog_data):
            """ Remove any old references to the catalog."""
            for key in {'catalog_dict', 'catalog_file'}:
                _ = catalog_data.pop(key, None)
            return catalog_data

        from pathlib import Path

        csv_file_name = Path(f'{name}.csv.gz')
        json_file_name = Path(f'{name}.json')
        if directory:
            directory = Path(directory)
            directory.mkdir(parents=True, exist_ok=True)
            csv_file_name = directory / csv_file_name
            json_file_name = directory / json_file_name

        collection_data = self.esmcol_data.copy()
        collection_data = _clear_old_catalog(collection_data)
        collection_data['id'] = name

        catalog_length = len(self.df)
        if catalog_type == 'file':
            collection_data['catalog_file'] = csv_file_name.as_posix()
            print(f'Writing csv catalog with {catalog_length} entries to: {csv_file_name}')
            self.df.to_csv(csv_file_name, compression='gzip', index=False)
        else:
            print(f'Writing catalog with {catalog_length} entries into: {json_file_name}')
            collection_data['catalog_dict'] = self.df.to_dict(orient='records')

        print(f'Writing ESM collection json file to: {json_file_name}')
        with open(json_file_name, 'w') as outfile:
            json.dump(collection_data, outfile)

    def nunique(self) -> pd.Series:
        """Count distinct observations across dataframe columns
        in the catalog.

        Examples
        --------
        >>> import intake
        >>> col = intake.open_esm_datastore("pangeo-cmip6.json")
        >>> col.nunique()
        activity_id          10
        institution_id       23
        source_id            48
        experiment_id        29
        member_id            86
        table_id             19
        variable_id         187
        grid_label            7
        zstore            27437
        dcpp_init_year       59
        dtype: int64
        """

        uniques = self.unique(self.df.columns.tolist())
        nuniques = {}
        for key, val in uniques.items():
            nuniques[key] = val['count']
        return pd.Series(nuniques)

    def unique(self, columns: Union[str, List] = None) -> Dict[str, Any]:
        """Return unique values for given columns in the
        catalog.

        Parameters
        ----------
        columns : str, list
           name of columns for which to get unique values

        Returns
        -------
        info : dict
           dictionary containing count, and unique values

        Examples
        --------
        >>> import intake
        >>> import pprint
        >>> col = intake.open_esm_datastore("pangeo-cmip6.json")
        >>> uniques = col.unique(columns=["activity_id", "source_id"])
        >>> pprint.pprint(uniques)
        {'activity_id': {'count': 10,
                        'values': ['AerChemMIP',
                                    'C4MIP',
                                    'CMIP',
                                    'DAMIP',
                                    'DCPP',
                                    'HighResMIP',
                                    'LUMIP',
                                    'OMIP',
                                    'PMIP',
                                    'ScenarioMIP']},
        'source_id': {'count': 17,
                    'values': ['BCC-ESM1',
                                'CNRM-ESM2-1',
                                'E3SM-1-0',
                                'MIROC6',
                                'HadGEM3-GC31-LL',
                                'MRI-ESM2-0',
                                'GISS-E2-1-G-CC',
                                'CESM2-WACCM',
                                'NorCPM1',
                                'GFDL-AM4',
                                'GFDL-CM4',
                                'NESM3',
                                'ECMWF-IFS-LR',
                                'IPSL-CM6A-ATM-HR',
                                'NICAM16-7S',
                                'GFDL-CM4C192',
                                'MPI-ESM1-2-HR']}}

        """
        return _unique(self.df, columns)

    def to_dataset_dict(
        self,
        zarr_kwargs: Dict[str, Any] = None,
        cdf_kwargs: Dict[str, Any] = None,
        preprocess: Dict[str, Any] = None,
        storage_options: Dict[str, Any] = None,
        progressbar: bool = None,
        aggregate: bool = None,
    ) -> Dict[str, xr.Dataset]:
        """
        Load catalog entries into a dictionary of xarray datasets.

        Parameters
        ----------
        zarr_kwargs : dict
            Keyword arguments to pass to :py:func:`~xarray.open_zarr` function
        cdf_kwargs : dict
            Keyword arguments to pass to :py:func:`~xarray.open_dataset` function.  If specifying chunks, the chunking
            is applied to each netcdf file.  Therefore, chunks must refer to dimensions that are present in each netcdf
            file, or chunking will fail.
        preprocess : callable, optional
            If provided, call this function on each dataset prior to aggregation.
        storage_options : dict, optional
            Parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        progressbar : bool
            If True, will print a progress bar to standard error (stderr)
            when loading assets into :py:class:`~xarray.Dataset`.
        aggregate : bool, optional
            If False, no aggregation will be done.

        Returns
        -------
        dsets : dict
           A dictionary of xarray :py:class:`~xarray.Dataset`.

        Examples
        --------
        >>> import intake
        >>> col = intake.open_esm_datastore("glade-cmip6.json")
        >>> cat = col.search(
        ...     source_id=["BCC-CSM2-MR", "CNRM-CM6-1", "CNRM-ESM2-1"],
        ...     experiment_id=["historical", "ssp585"],
        ...     variable_id="pr",
        ...     table_id="Amon",
        ...     grid_label="gn",
        ... )
        >>> dsets = cat.to_dataset_dict()
        >>> dsets.keys()
        dict_keys(['CMIP.BCC.BCC-CSM2-MR.historical.Amon.gn', 'ScenarioMIP.BCC.BCC-CSM2-MR.ssp585.Amon.gn'])
        >>> dsets["CMIP.BCC.BCC-CSM2-MR.historical.Amon.gn"]
        <xarray.Dataset>
        Dimensions:    (bnds: 2, lat: 160, lon: 320, member_id: 3, time: 1980)
        Coordinates:
        * lon        (lon) float64 0.0 1.125 2.25 3.375 ... 355.5 356.6 357.8 358.9
        * lat        (lat) float64 -89.14 -88.03 -86.91 -85.79 ... 86.91 88.03 89.14
        * time       (time) object 1850-01-16 12:00:00 ... 2014-12-16 12:00:00
        * member_id  (member_id) <U8 'r1i1p1f1' 'r2i1p1f1' 'r3i1p1f1'
        Dimensions without coordinates: bnds
        Data variables:
            lat_bnds   (lat, bnds) float64 dask.array<chunksize=(160, 2), meta=np.ndarray>
            lon_bnds   (lon, bnds) float64 dask.array<chunksize=(320, 2), meta=np.ndarray>
            time_bnds  (time, bnds) object dask.array<chunksize=(1980, 2), meta=np.ndarray>
            pr         (member_id, time, lat, lon) float32 dask.array<chunksize=(1, 600, 160, 320), meta=np.ndarray>
        """

        # Return fast
        if not self.keys():
            warn('There are no datasets to load! Returning an empty dictionary.')
            return {}

        source_kwargs = OrderedDict(
            zarr_kwargs=zarr_kwargs,
            cdf_kwargs=cdf_kwargs,
            preprocess=preprocess,
            storage_options=storage_options,
        )

        if progressbar is not None:
            self.progressbar = progressbar

        if preprocess is not None and not callable(preprocess):
            raise ValueError('preprocess argument must be callable')

        if aggregate is not None and not aggregate:
            self = deepcopy(self)
            self.groupby_attrs = []

        if self.progressbar:
            print(
                f"""\n--> The keys in the returned dictionary of datasets are constructed as follows:\n\t'{self.key_template}'"""
            )

        def _load_source(key, source):
            return key, source.to_dask()

        sources = {key: source(**source_kwargs) for key, source in self.items()}
        progress, total = None, None
        if self.progressbar:
            total = len(sources)
            progress = progress_bar(range(total))

        self._datasets = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=dask.system.CPU_COUNT) as executor:
            future_tasks = [
                executor.submit(_load_source, key, source) for key, source in sources.items()
            ]
            for i, task in enumerate(concurrent.futures.as_completed(future_tasks)):
                key, ds = task.result()
                self._datasets[key] = ds
                if self.progressbar:
                    progress.update(i)
            if self.progressbar:
                progress.update(total)
            return self._datasets


def _make_entry(key: str, driver: str, args: dict):
    entry = intake.catalog.local.LocalCatalogEntry(
        name=key, description='', driver=driver, args=args, metadata={}
    )
    return entry.get()


def _construct_agg_info(aggregations: List[Dict]) -> Tuple[List[Dict], Dict, List]:
    """
    Helper function used to determine aggregation columns information and their
    respective settings.

    Examples
    --------

    >>> a = [
    ...     {"type": "union", "attribute_name": "variable_id"},
    ...     {
    ...         "type": "join_new",
    ...         "attribute_name": "member_id",
    ...         "options": {"coords": "minimal", "compat": "override"},
    ...     },
    ...     {
    ...         "type": "join_new",
    ...         "attribute_name": "dcpp_init_year",
    ...         "options": {"coords": "minimal", "compat": "override"},
    ...     },
    ... ]
    >>> aggregations, aggregation_dict, agg_columns = _construct_agg_info(a)
    >>> agg_columns
    ['variable_id', 'member_id', 'dcpp_init_year']
    >>> aggregation_dict
    {'variable_id': {'type': 'union'},
    'member_id': {'type': 'join_new',
    'options': {'coords': 'minimal', 'compat': 'override'}},
    'dcpp_init_year': {'type': 'join_new',
    'options': {'coords': 'minimal', 'compat': 'override'}}}
    """
    agg_columns = []
    aggregation_dict = {}
    if aggregations:
        # Sort aggregations to make sure join_existing is always done before join_new
        aggregations = sorted(aggregations, key=lambda i: i['type'], reverse=True)
        for agg in aggregations:
            key = agg['attribute_name']
            if agg['type'] == 'join_existing' and 'dim' not in agg['options']:
                message = f"""
            Missing `dim` option for `join_existing` operation across `{key}` attribute.
            For `join_existing` to properly work, `options` must contain the name of the existing dimension
            to use (for e.g.: something like {{'dim': 'time'}}).
                """
                warn(message)
            rest = agg.copy()
            del rest['attribute_name']
            aggregation_dict[key] = rest
        agg_columns = list(aggregation_dict.keys())
    return aggregations, aggregation_dict, agg_columns



""" Functions for aggregating multiple xarray datasets into a single xarray dataset"""
from typing import Any, Dict, List, Union

import fsspec
import xarray as xr

from .search import _flatten_list


class AggregationError(Exception):
    pass


def _path_to_mapper(path, storage_options, data_format=None):
    """Convert path to mapper if necessary."""
    if fsspec.core.split_protocol(path)[0] is not None:
        if data_format == 'netcdf':
            return fsspec.open(path, **storage_options)
        return fsspec.get_mapper(path, **storage_options)
    return path


def join_new(
    dsets: List[xr.Dataset],
    dim_name: str,
    coord_value: Any,
    varname: Union[str, List[str]],
    options: Dict[str, Any] = None,
    group_key: str = None,
) -> xr.Dataset:
    """
    Concatenate a list of datasets along a new dimension.

    Parameters
    ----------
    dsets : List[xr.Dataset]
        A list of xarray.Dataset(s) to concatenate along a new dimension
    dim_name : str
        Name of the new dimension
    coord_value : Any
        Value corresponding to the new dimension coordinate
    varname : List
        List of data variables
    options : Dict, optional
        Additional keyword arguments passed through to
        :py:func:`~xarray.concat()`, by default None

    Returns
    -------
    xr.Dataset
        xarray Dataset
    """
    options = options or {}
    try:
        concat_dim = xr.DataArray(coord_value, dims=(dim_name), name=dim_name)
        return xr.concat(dsets, dim=concat_dim, data_vars=varname, **options)
    except Exception as exc:
        message = f"""
        Failed to join/concatenate datasets in group with key={group_key} along a new dimension `{dim_name}`.

        *** Arguments passed to xarray.concat() ***:

        - objs: a list of {len(dsets)} datasets
        - dim: {concat_dim}
        - data_vars: {varname}
        - and kwargs: {options}

        ********************************************
        """

        raise AggregationError(message) from exc


def join_existing(
    dsets: List[xr.Dataset],
    varname: Union[str, List[str]],
    options: Dict[str, Any] = None,
    group_key: str = None,
) -> xr.Dataset:
    """
    Concatenate a list of datasets along an existing dimension.

    Parameters
    ----------
    dsets : List[xr.Dataset]
        A list of xarray.Dataset(s) to concatenate along an existing dimension.
    varname : List
        List of data variables
    options : Dict, optional
        Additional keyword arguments passed through to
        :py:func:`~xarray.concat()`, by default None

    Returns
    -------
    xr.Dataset
        xarray Dataset
    """
    options = options or {}
    try:
        return xr.concat(dsets, data_vars=varname, **options)
    except Exception as exc:
        message = f"""
        Failed to join/concatenate datasets in group with key={group_key} along an existing dimension.

        *** Arguments passed to xarray.concat() ***:

        - objs: a list of {len(dsets)} datasets
        - data_vars: {varname}
        - kwargs: {options}

        ********************************************
        """
        raise AggregationError(message) from exc


def union(
    dsets: List[xr.Dataset], options: Dict[str, Any] = None, group_key: str = None
) -> xr.Dataset:
    """
    Merge a list of datasets into a single dataset.

    Parameters
    ----------
    dsets : List[xr.Dataset]
        A list of xarray.Dataset(s) to merge.
    options : Dict, optional
        Additional keyword arguments passed through to
        :py:func:`~xarray.merge()`, by default None

    Returns
    -------
    xr.Dataset
        xarray Dataset
    """

    options = options or {}
    try:
        return xr.merge(dsets, **options)
    except Exception as exc:
        message = f"""
        Failed to merge multiple datasets in group with key={group_key} into a single xarray Dataset as variables.

        *** Arguments passed to xarray.merge() ***:

        - objs: a list of {len(dsets)} datasets
        - kwargs: {options}

        ********************************************
        """
        raise AggregationError(message) from exc


def _to_nested_dict(df):
    """Converts a multiindex series to nested dict"""
    if hasattr(df.index, 'levels') and len(df.index.levels) > 1:
        ret = {}
        for k, v in df.groupby(level=0):
            ret[k] = _to_nested_dict(v.droplevel(0))
        return ret
    return df.to_dict()


def _aggregate(
    aggregation_dict: Dict[str, Any],
    agg_columns: List[str],
    n_agg: Dict[str, Any],
    nd: Dict[str, Any],
    mapper_dict: Dict[str, Any],
    group_key: str = None,
):
    def apply_aggregation(nd, agg_column=None, key=None, level=0):
        """Recursively descend into nested dictionary and aggregate items.
        level tells how deep we are."""

        assert level <= n_agg

        if level == n_agg:
            # bottom of the hierarchy - should be an actual dataset
            # return dataset at this point
            ds = mapper_dict[nd]
            if isinstance(ds, xr.Dataset):
                return ds
            raise TypeError(
                f'Expected mapper_dict[{nd}] to be an xarray.Dataset. Found type of mapper_dict[{nd}] to be {type(mapper_dict[nd])}'
            )

        agg_column = agg_columns[level]
        agg_info = aggregation_dict[agg_column]
        agg_type = agg_info['type']

        if 'options' in agg_info:
            agg_options = agg_info['options']
        else:
            agg_options = {}

        dsets = [
            apply_aggregation(value, agg_column, key=key, level=level + 1)
            for key, value in nd.items()
        ]
        keys = list(nd.keys())
        attrs = dict_union(*[ds.attrs for ds in dsets])
        # copy encoding for each variable from first encounter
        variables = {v for ds in dsets for v in ds.variables}
        encoding = {}
        for ds in dsets:
            for v in variables:
                if v in ds.variables and v not in encoding:
                    if ds[v].encoding:
                        encoding[v] = ds[v].encoding
                        # get rid of the misleading file-specific attributes
                        # github.com/pydata/xarray/issues/2550
                        for enc_attrs in ['source', 'original_shape']:
                            if enc_attrs in encoding[v]:
                                del encoding[v][enc_attrs]

        if agg_type == 'join_new':
            varname = dsets[0].attrs['intake_esm_varname']
            ds = join_new(
                dsets,
                dim_name=agg_column,
                coord_value=keys,
                varname=varname,
                options=agg_options,
                group_key=group_key,
            )

        elif agg_type == 'join_existing':
            varname = dsets[0].attrs['intake_esm_varname']
            ds = join_existing(dsets, varname=varname, options=agg_options, group_key=group_key)

        elif agg_type == 'union':
            ds = union(dsets, options=agg_options, group_key=group_key)

        ds.attrs = attrs
        for v in ds.variables:
            if v in encoding and not ds[v].encoding:
                ds[v].encoding = encoding[v]
        return ds

    return apply_aggregation(nd)


def _open_asset(
    path,
    data_format,
    zarr_kwargs=None,
    cdf_kwargs=None,
    preprocess=None,
    varname: Union[List[str], str] = None,
    requested_variables: Union[List[str], str] = None,
):
    def normalize_protocol(protocol):
        if isinstance(protocol, list):
            return tuple(protocol)
        return protocol

    protocol, root = None, path
    if isinstance(path, fsspec.mapping.FSMap):
        protocol = normalize_protocol(path.fs.protocol)
        if protocol in {'http', 'https', 'file'} or protocol is None:
            path = path.root
            root = path
        else:
            root = path.root

    elif isinstance(path, fsspec.core.OpenFile):
        protocol = normalize_protocol(path.fs.protocol)
        root = path.path
        path = path.open()

    if data_format == 'zarr':
        try:
            ds = xr.open_zarr(path, **zarr_kwargs)
        except Exception as exc:
            message = f"""
            Failed to open zarr store.

            *** Arguments passed to xarray.open_zarr() ***:

            - store: {path}
            - kwargs: {zarr_kwargs}

            *** fsspec options used ***:

            - root: {root}
            - protocol: {protocol}

            ********************************************
            """

            raise IOError(message) from exc

    else:
        try:
            ds = xr.open_dataset(path, **cdf_kwargs)
        except Exception as exc:
            message = f"""
            Failed to open netCDF/HDF dataset.

            *** Arguments passed to xarray.open_dataset() ***:

            - filename_or_obj: {path}
            - kwargs: {cdf_kwargs}

            *** fsspec options used ***:

            - root: {root}
            - protocol: {protocol}

            ********************************************
            """
            raise IOError(message) from exc
    if preprocess is not None:
        try:
            ds = preprocess(ds)
        except Exception as exc:
            raise RuntimeError(
                f'Failed to apply pre-processing function: {preprocess.__name__}'
            ) from exc

    if varname:
        if isinstance(varname, str):
            varname = [varname]

    if requested_variables:
        if isinstance(requested_variables, str):
            requested_variables = [requested_variables]
        variable_intersection = set(requested_variables).intersection(set(varname))
        variables = [variable for variable in variable_intersection if variable in ds.data_vars]
        ds = ds[variables]
        ds.attrs['intake_esm_varname'] = variables

    else:
        ds.attrs['intake_esm_varname'] = varname

    return ds


def dict_union(*dicts, merge_keys=['history', 'tracking_id', 'intake_esm_varname'], drop_keys=[]):
    """Return the union of two or more dictionaries."""
    from functools import reduce

    if len(dicts) > 2:
        return reduce(dict_union, dicts)
    elif len(dicts) == 2:
        d1, d2 = dicts
        d = type(d1)()
        # union
        all_keys = set(d1) | set(d2)
        for k in all_keys:
            v1 = d1.get(k)
            v2 = d2.get(k)
            if (v1 is None and v2 is None) or k in drop_keys:
                pass
            elif v1 is None:
                d[k] = v2
            elif v2 is None:
                d[k] = v1
            elif v1 == v2:
                d[k] = v1
            elif k in merge_keys:
                d[k] = '\n'.join(_flatten_list([v1, v2]))
        return d
    elif len(dicts) == 1:
        return dicts[0]



# -*- coding: utf-8 -*-
""" Helper functions for fetching and loading catalog"""
import importlib
import json
import sys
from pathlib import Path
from urllib.parse import ParseResult, urlparse, urlunparse

import pandas as pd
import requests


def _is_valid_url(url):
    """Check if path is URL or not
    Parameters
    ----------
    url : str
        path to check
    Returns
    -------
    bool
    """
    try:
        result = urlparse(url)
        return (
            result.scheme
            and result.netloc
            and result.path
            and (requests.get(url).status_code == 200)
        )
    except Exception:
        return False


def _fetch_and_parse_json(input_path):
    """Fetch and parse ESMCol file.
    Parameters
    ----------
    input_path : str
            ESMCol file to get and read
    Returns
    -------
    data : dict
    input_path : str
    """

    data = None

    try:
        if _is_valid_url(input_path):
            resp = requests.get(input_path)
            data = resp.json()
        else:
            input_path = Path(input_path).absolute().as_posix()
            with open(input_path) as filein:
                data = json.load(filein)

    except Exception as exc:
        raise exc

    return data, input_path


def _fetch_catalog(collection_data, esmcol_path, csv_kwargs=None):
    """Get the catalog file content, and load it into a pandas dataframe"""

    if csv_kwargs is None:
        csv_kwargs = {}
    catalog_path = None
    if 'catalog_file' in collection_data:
        if _is_valid_url(esmcol_path):
            catalog_path = collection_data['catalog_file']
            if not _is_valid_url(catalog_path):
                split_url = urlparse(esmcol_path)
                path = (Path(split_url.path).parent / collection_data['catalog_file']).as_posix()
                components = ParseResult(
                    scheme=split_url.scheme,
                    netloc=split_url.netloc,
                    path=path,
                    params=split_url.params,
                    query=split_url.query,
                    fragment=split_url.fragment,
                )
                catalog_path = urlunparse(components)
                if not _is_valid_url(catalog_path):
                    raise FileNotFoundError(f'Unable to find: {catalog_path}')
        else:
            catalog_path = Path(collection_data['catalog_file'])
            # If the catalog_path does not exist,
            # try constructing a path using the relative path
            if not catalog_path.exists():
                esmcol_path = Path(esmcol_path).absolute()
                catalog_path = esmcol_path.parent / collection_data['catalog_file']
                if not catalog_path.exists():
                    raise FileNotFoundError(f'Unable to find: {catalog_path}')
        return pd.read_csv(catalog_path, **csv_kwargs), catalog_path
    return pd.DataFrame(collection_data['catalog_dict']), None


def show_versions(file=sys.stdout):  # pragma: no cover
    """print the versions of intake-esm and its dependencies.
       Adapted from xarray/util/print_versions.py

    Parameters
    ----------
    file : file-like, optional
        print to the given file-like object. Defaults to sys.stdout.
    """

    deps = [
        ('xarray', lambda mod: mod.__version__),
        ('pandas', lambda mod: mod.__version__),
        ('intake', lambda mod: mod.__version__),
        ('intake_esm', lambda mod: mod.__version__),
        ('fsspec', lambda mod: mod.__version__),
        ('s3fs', lambda mod: mod.__version__),
        ('gcsfs', lambda mod: mod.__version__),
        ('fastprogress', lambda mod: mod.__version__),
        ('dask', lambda mod: mod.__version__),
        ('zarr', lambda mod: mod.__version__),
        ('cftime', lambda mod: mod.__version__),
        ('netCDF4', lambda mod: mod.__version__),
        ('requests', lambda mod: mod.__version__),
    ]

    deps_blob = []
    for (modname, ver_f) in deps:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = importlib.import_module(modname)
        except Exception:
            deps_blob.append((modname, None))
        else:
            try:
                ver = ver_f(mod)
                deps_blob.append((modname, ver))
            except Exception:
                deps_blob.append((modname, 'installed'))

    print('\nINSTALLED VERSIONS', file=file)
    print('------------------', file=file)

    print('', file=file)
    for k, stat in sorted(deps_blob):
        print(f'{k}: {stat}', file=file)



# -*- coding: utf-8 -*-
#
# complexity documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
import sys

import yaml

import intake_esm

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

cwd = os.getcwd()
parent = os.path.dirname(cwd)
sys.path.insert(0, parent)


# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'IPython.sphinxext.ipython_console_highlighting',
    'IPython.sphinxext.ipython_directive',
    'sphinx.ext.napoleon',
    'myst_nb',
]

autodoc_member_order = 'groupwise'

myst_amsmath_enable = True
myst_admonition_enable = True
myst_html_img_enable = True
myst_dmath_enable = True
myst_deflist_enable = True
myst_figure_enable = True
myst_url_schemes = ('http', 'https', 'mailto')
myst_heading_anchors = 2
panels_add_boostrap_css = False

jupyter_execute_notebooks = 'cache'
# jupyter_execute_notebooks = 'off'
execution_timeout = 60

extlinks = {
    'issue': ('https://github.com/intake/intake-esm/issues/%s', 'GH#'),
    'pr': ('https://github.com/intake/intake-esm/pull/%s', 'GH#'),
}
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Autosummary pages will be generated by sphinx-autogen instead of sphinx-build
autosummary_generate = []

# Otherwise, the Return parameter list looks different from the Parameters list
napoleon_use_rtype = False


# The suffix of source filenames.
# source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
current_year = datetime.datetime.now().year
project = u'Intake-ESM'
copyright = f'2018-{current_year}, Intake-ESM development team'
author = u'Intake-ESM developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = intake_esm.__version__.split('+')[0]
# The full version, including alpha/beta/rc tags.
release = intake_esm.__version__


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**.ipynb_checkpoints', 'Thumbs.db', '.DS_Store']


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'pydata_sphinx_theme'


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'images/NSF_4-Color_bitmap_Logo.png'
html_context = {
    'github_user': 'intake',
    'github_repo': 'intake-esm',
    'github_version': 'master',
    'doc_path': 'docs',
}
html_theme_options = {
    'github_url': 'https://github.com/intake/intake-esm',
    'twitter_url': 'https://twitter.com/NCARXDev',
    'show_toc_level': 1,
}


# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_build/html/_static']

# Sometimes the savefig directory doesn't exist and needs to be created
# https://github.com/ipython/ipython/issues/8733
# becomes obsolete when we can pin ipython>=5.2; see ci/requirements/doc.yml
ipython_savefig_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '_build', 'html', '_static'
)

savefig_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'source', '_static')

os.makedirs(ipython_savefig_dir, exist_ok=True)
os.makedirs(savefig_dir, exist_ok=True)

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'


# Output file base name for HTML help builder.
htmlhelp_basename = 'intake_esmdoc'


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}


latex_documents = [('index', 'intake-esm.tex', u'intake-esm Documentation', author, 'manual')]

man_pages = [('index', 'intake-esm', u'intake-esm Documentation', [author], 1)]

texinfo_documents = [
    (
        'index',
        'intake-esm',
        u'intake-esm Documentation',
        author,
        'intake-esm',
        'One line description of project.',
        'Miscellaneous',
    )
]

ipython_warning_is_error = False
ipython_execlines = [
    'import intake',
    'import intake_esm',
    'import xarray',
    'import pandas as pd',
    'pd.options.display.encoding="utf8"',
]


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'xarray': ('http://xarray.pydata.org/en/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'intake': ('https://intake.readthedocs.io/en/latest/', None),
}


# https://www.ericholscher.com/blog/2016/jul/25/integrating-jinja-rst-sphinx/


def rstjinja(app, docname, source):
    """
    Render our pages as a jinja template for fancy templating goodness.
    """
    # Make sure we're outputting HTML
    if app.builder.format != 'html':
        return
    src = source[0]
    rendered = app.builder.templates.render_string(src, app.config.html_context)
    source[0] = rendered


def setup(app):
    app.connect('source-read', rstjinja)
    # app.add_stylesheet('style.css')


with open('catalogs.yaml') as f:
    catalogs = yaml.safe_load(f)


html_context = {'catalogs': catalogs['catalogs']}



import itertools
from collections.abc import Iterable
from typing import Pattern
from warnings import warn

import numpy as np
import pandas as pd


def _unique(df, columns=None):
    if isinstance(columns, str):
        columns = [columns]
    if not columns:
        columns = df.columns.tolist()
    info = {}
    for col in columns:
        values = df[col].dropna().values
        uniques = np.unique(list(_flatten_list(values))).tolist()
        info[col] = {'count': len(uniques), 'values': uniques}
    return info


def search(df, require_all_on=None, **query):
    """
    Search for entries in a pandas DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame to run query against.
    require_all_on : list, str, optional
        A dataframe column or a list of dataframe columns across
        which all entries must satisfy the query criteria.
        If None, return entries that fulfill any of the criteria specified
        in the query, by default None.
    **query:
        keyword arguments corresponding to user's query to execute against the dataframe.

    Returns
    -------
    pd.DataFrame
    """
    columns_with_iterables = _get_columns_with_iterables(df)
    message = 'Query returned zero results.'
    if not query:
        warn(message)
        return pd.DataFrame(columns=df.columns)
    condition = np.ones(len(df), dtype=bool)
    query = _normalize_query(query)
    for key, val in query.items():
        condition_i = np.zeros(len(df), dtype=bool)
        column_is_stringtype = isinstance(
            df[key].dtype, (np.object, pd.core.arrays.string_.StringDtype)
        )
        column_has_iterables = key in columns_with_iterables
        for val_i in val:
            if column_has_iterables:
                cond = df[key].str.contains(val_i, regex=False)
            else:
                value_is_pattern = _is_pattern(val_i)
                if column_is_stringtype and value_is_pattern:
                    cond = df[key].str.contains(val_i, regex=True, case=True, flags=0)
                else:
                    cond = df[key] == val_i
            condition_i = condition_i | cond
        condition = condition & condition_i
    query_results = df.loc[condition]

    if require_all_on:
        if isinstance(require_all_on, str):
            require_all_on = [require_all_on]
        _query = query.copy()

        # Make sure to remove columns that were already
        # specified in the query when specified in `require_all_on`. For example,
        # if query = dict(variable_id=["A", "B"], source_id=["FOO", "BAR"])
        # and require_all_on = ["source_id"], we need to make sure `source_id` key is
        # not present in _query for the logic below to work
        for key in require_all_on:
            _query.pop(key, None)

        keys = list(_query.keys())
        grouped = query_results.groupby(require_all_on)
        values = [tuple(v) for v in _query.values()]
        condition = set(itertools.product(*values))
        results = []
        for key, group in grouped:
            index = group.set_index(keys).index
            if not isinstance(index, pd.MultiIndex):
                index = {(element,) for element in index.to_list()}
            else:
                index = set(index.to_list())
            if index == condition:
                results.append(group)

        if len(results) >= 1:
            return pd.concat(results).reset_index(drop=True)

        warn(message)
        return pd.DataFrame(columns=df.columns)
    if query_results.empty:
        warn(message)

    return query_results.reset_index(drop=True)


def _normalize_query(query):
    q = query.copy()
    for key, val in q.items():
        if isinstance(val, str) or not isinstance(val, Iterable):
            q[key] = [val]
    return q


def _is_pattern(value):
    if isinstance(value, Pattern):
        return True
    wildcard_chars = {'*', '?', '$', '^'}
    try:
        value_ = value
        for char in wildcard_chars:
            value_ = value_.replace(fr'\{char}', '')
        return any(char in value_ for char in wildcard_chars)
    except (TypeError, AttributeError):
        return False


def _flatten_list(data):
    for item in data:
        if isinstance(item, Iterable) and not isinstance(item, str):
            for x in _flatten_list(item):
                yield x
        else:
            yield item


def _get_columns_with_iterables(df):
    if not df.empty:
        has_iterables = (
            df.sample(20, replace=True).applymap(type).isin([list, tuple, set]).any().to_dict()
        )
        columns_with_iterables = [column for column, check in has_iterables.items() if check]
    else:
        columns_with_iterables = []
    return columns_with_iterables
